/*
    Copyright 2009,2010 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/sha.h>
#include <openssl/rc4.h>
#include <windows.h>    // Windows only, I have not checked the Linux and Mac one

typedef unsigned char   u8;



#define VER     "0.2"



int unbase64(u8 *in, int insz, u8 *out, int outsz) {
    int     xlen,
            a   = 0,
            b   = 0,
            c,
            step;
    u8      *limit,
            *data,
            *p;
    static const u8 base[128] = {   // supports also the Gamespy base64 and URLs
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3e,0x00,0x3e,0x00,0x3f,
        0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x00,0x00,0x00,0x00,0x00,0x00,
        0x00,0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,
        0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x3e,0x00,0x3f,0x00,0x3f,
        0x00,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,
        0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,0x33,0x00,0x00,0x00,0x00,0x00
    };

    if(insz < 0) insz = strlen(in);
    xlen = ((insz >> 2) * 3) + 1;    // NULL included in output for text
    if((in != out) && (outsz >= 0)) {
        if(outsz < xlen) return(-1);
    } else {
        outsz = xlen;
    }
    data = in;

    p = out;
    limit = data + insz;

    for(step = 0; /* data < limit */; step++) {
        do {
            if(data >= limit) {
                c = 0;
                break;
            }
            c = *data;
            data++;
            if((c == '=') || (c == '_')) {  // supports also the Gamespy base64
                c = 0;
                break;
            }
        } while(c && ((c <= ' ') || (c > 0x7f)));
        if(!c) break;

        switch(step & 3) {
            case 0: {
                a = base[c];
                break;
            }
            case 1: {
                b = base[c];
                *p++ = (a << 2)        | (b >> 4);
                break;
            }
            case 2: {
                a = base[c];
                *p++ = ((b & 15) << 4) | (a >> 2);
                break;
            }
            case 3: {
                *p++ = ((a & 3) << 6)  | base[c];
                break;
            }
        }
    }
    *p = 0;
    return(p - out);
}



u8 *regkeyr(HKEY hKey, LPCTSTR lpSubKey, LPTSTR lpValueName, int *ret_len) {
    HKEY    key     = NULL;
    DWORD   len     = 0;
    u8      *buff   = NULL;

    if(ret_len) *ret_len = len;

    if(RegOpenKeyEx(hKey, lpSubKey, 0, KEY_READ, &key) != ERROR_SUCCESS)  goto quit;

    if(RegQueryValueEx(key, lpValueName, NULL, REG_NONE, NULL, &len) != ERROR_SUCCESS)  goto quit;

    buff = malloc(len + 1);
    if(!buff) goto quit;

    if(RegQueryValueEx(key, lpValueName, NULL, REG_NONE, buff, &len) != ERROR_SUCCESS) goto quit;

    buff[len] = 0;
    if(ret_len) *ret_len = len;
    RegCloseKey(key);
    return(buff);
quit:
    if(key) RegCloseKey(key);
    return(NULL);
}



int cstring(u8 *str) {
    int     c,
            n;
    u8      *p,
            *o;

    p = strchr(str, '\'');
    if(!p) p = strchr(str, '\"');
    if(p) {
        p++;
    } else {
        p = str;
    }
    for(o = str; *p; o++) {
        if(*p == '\'') break;
        if(*p == '\"') break;
        n = 1;
        if(*p == '\\') {
            p++;
            switch(*p) {
                case 0:    c = 0;    break;
                case '0':  c = '\0'; break;
                case 'a':  c = '\a'; break;
                case 'b':  c = '\b'; break;
                case 'e':  c = '\e'; break;
                case 'f':  c = '\f'; break;
                case 'n':  c = '\n'; break;
                case 'r':  c = '\r'; break;
                case 't':  c = '\t'; break;
                case 'v':  c = '\v'; break;
                case '\"': c = '\"'; break;
                case '\'': c = '\''; break;
                case '\\': c = '\\'; break;
                case '?':  c = '\?'; break;
                case '.':  c = '.';  break;
                case ' ':  c = ' ';  break;
                case 'x':  sscanf(++p, "%02x%n", &c, &n);   break;
                default:   sscanf(p, "%03o%n", &c, &n);     break;
            }
        } else {
            c = *p;
        }
        *o = c;
        p += n;
    }
    *o = 0;
    return(o - str);
}



void digsby_pwd(u8 *user, u8 *pass, int passlen) {
    RC4_KEY ctx;
    int     pidlen,
            datelen,
            userlen,
            keysz;
    u8      date[32],
            *pid    = NULL,
            *tmp    = NULL,
            *key    = NULL;

    if(user) {
        userlen = strlen(user);
        pid = regkeyr(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion", "DigitalProductId", &pidlen);
        tmp = regkeyr(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion", "InstallDate", NULL);
        datelen = 0;

        if(tmp) datelen = sprintf(date, "%u", *(int *)tmp);
        keysz = pidlen + datelen + userlen;
        key = malloc(keysz);
        memcpy(key, pid, pidlen);
        memcpy(key + pidlen, date, datelen);
        memcpy(key + pidlen + datelen, user, userlen);

        SHA1(key, keysz, key);
        keysz = SHA_DIGEST_LENGTH;
    }
    if(!user) { // separated for possible future implementations
        keysz = 3;
        key = malloc(keysz);
        memcpy(key, "foo", keysz);
    }

    RC4_set_key(&ctx, keysz, key);
    RC4(&ctx, passlen, pass, pass);

    if(pid) free(pid);
    if(tmp) free(tmp);
    if(key) free(key);
}



int fgetz(u8 *buff, int buffsz, FILE *fd) {
    u8      *p;

    if(!fgets(buff, buffsz, fd)) return(0);
    for(p = buff; *p && (*p != '\r') && (*p != '\n'); p++);
    *p = 0;
    return(1);
}



#define BORING(X,Y) \
                if(*l == '!') { \
                    if(!strnicmp(l + 1, "binary", 6)) { \
                        boring = Y; \
                        b64 = 1; \
                    } else { \
                        while(*l && (*l > ' ')) l++; \
                        if(*l) { \
                            l++; \
                            while(*l && (*l <= ' ')) l++; \
                            X = strdup(l); \
                        } \
                    } \
                } else { \
                    X = strdup(l); \
                }
#define BORING_RESET \
                if(user) free(user); \
                if(pass) free(pass); \
                user = NULL; \
                pass = NULL; \
                b64  = 0;



int main(int argc, char *argv[]) {
    FILE    *fd;
    int     boring  = 0,
            b64     = 0,
            passlen,
            oldver  = 0;
    u8      buff[4096],
            *pass   = NULL,
            *user   = NULL,
            *p,
            *l;

    fputs("\n"
        "Digsby password decoder "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <digsby.dat/logininfo.yaml/password> [username]\n"
            "\n"
            " logininfo.yaml is located in:\n"
            "  \"%%USERPROFILE%%\\Local Settings\\Application Data\\Digsby\\logininfo.yaml\"\n"
            "\n"
            " digsby.dat was used in old versions of Digsby:\n"
            "  %%APPDATA%%\\Digsby\\digsby.dat\n"
            "\n"
            " it's also possible to specify directly the password but it's required to\n"
            " specify also the corresponding username because it's used for the\n"
            " decryption (only the old digsby doesn't need it)\n"
            " \n"
            " examples: digsbypwd c:\\logininfo.yaml\n"
            "           digsbypwd bXlwYXNzd29yZA== myusername\n"
            //"examples: digsbypwd S'\\xc6[\\x9dH\\xbb\\x12,]D\\xee'\n"
            //"          digsbypwd \\xc6[\\x9dH\\xbb\\x12,]D\\xee\n"
            "\n", argv[0]);
        goto quit;
    }

    pass = argv[1];
    if(argc > 2) user = argv[2];

    fd = fopen(pass, "rb");
    if(fd) {
        if(stristr(pass, "digsby.dat")) oldver = 1;
        // the following code sux but works and I have written it on the fly
        user = NULL;
        pass = NULL;
        for(;;) {
            if(user && pass) {  // decrypt the password, must be the first of the cicle
                passlen = cstring(pass);
                if(b64) passlen = unbase64(pass, passlen, pass, passlen);
                digsby_pwd(oldver ? NULL : user, pass, passlen);
                printf("  Username: %s\n", user);
                printf("  Password: %s\n", pass);
                printf("\n");
                BORING_RESET
                // "continue;" is not necessary
            }

            if(!fgetz(buff, sizeof(buff), fd)) break;
            p = buff;
            while(*p && (*p <= ' ')) p++;
            if(!*p) continue;

            if(boring) {    // the binary mode uses the subsequent line
                switch(boring) {
                    case 1: user = strdup(p);   break;
                    case 2: pass = strdup(p);   break;
                    default: break;
                }
                boring = 0;
                continue;
            }

            l = strchr(p, ':');
            if(!l) continue;
            *l++ = 0;
            while(*l && (*l <= ' ')) l++;

            if(!stricmp(p, "?")) {
                BORING_RESET
            } else if(!stricmp(p, "username")) {
                BORING(user, 1)
            } else if(!stricmp(p, "password")) {
                BORING(pass, 2)
            }
        }
        fclose(fd);
    } else {
        passlen = cstring(pass);
        if(user) passlen = unbase64(pass, passlen, pass, passlen);  // new mode
        digsby_pwd(user, pass, passlen);
        printf("  Password: %s\n", pass);
        if(!user) printf("\n- the password has been decrypted using the old method\n  if it's incorrent you MUST specify also the username\n");
    }

quit:
    printf("\n  Press RETURN to exit\n");
    fgetc(stdin);
    return(0);
}


